@@ -12,26 +12,28 @@ from cyclotron.debug import trace_observable
 from .asyncio import to_agen
 
 from kafka.partitioner import murmur2
+import aiokafka
 from aiokafka import AIOKafkaConsumer, AIOKafkaProducer
 
 Sink = namedtuple('Sink', ['request'])
 Source = namedtuple('Source', ['response', 'feedback'])
 
 # Sink items
-Consumer = namedtuple('Consumer', ['server', 'topics'])
+Consumer = namedtuple('Consumer', ['server', 'topics', 'group', 'max_partition_fetch_bytes'])
 Consumer.__doc__ += ": Creates a consumer client that can subscribe to multiple topics"
 Consumer.server.__doc__ += ": Address of the boostrap server"
 Consumer.topics.__doc__ += ": Observable emitting ConsumerTopic items"
+Consumer.__new__.__defaults__ = (1048576,)
 
-Producer = namedtuple('Producer', ['server', 'topics', 'acks'])
-Producer.__new__.__defaults__ = (1,)
+Producer = namedtuple('Producer', ['server', 'topics', 'acks', 'max_request_size'])
+Producer.__new__.__defaults__ = (1, 1048576)
 Producer.__doc__ += ": Creates a producer client that can publish to pultiple topics"
 Producer.server.__doc__ += ": Address of the boostrap server"
 Producer.topics.__doc__ += ": Observable emitting ProducerTopic items"
 Producer.acks.__doc__ += ": Records acknowledgement strategy, as documented in aiokafka"
 
-ConsumerTopic = namedtuple('ConsumerTopic', ['topic', 'group', 'decode', 'control'])
-ConsumerTopic.__new__.__defaults__ = (None,)
+ConsumerTopic = namedtuple('ConsumerTopic', ['topic', 'decode', 'control', 'start_from'])
+ConsumerTopic.__new__.__defaults__ = (None, 'end')
 ProducerTopic = namedtuple('ProducerTopic', ['topic', 'records', 'map_key', 'encode', 'map_partition'])
 
 # Source items
@@ -65,12 +67,19 @@ async def send_record(client, topic, key, value, partition_key):
         )
 
 
-def run_consumer(loop, source_observer, server, topics):
+DelConsumerCmd = namedtuple('DelConsumerCmd', ['observer'])
+AddConsumerCmd = namedtuple('AddConsumerCmd', ['observer', 'consumer'])
+ConsumerProperties = namedtuple('ConsumerProperties', [
+    'observer', 'topic', 'decode', 'start_from'
+])
+
+
+def run_consumer(loop, source_observer, server, group, topics):
     async def _run_consumer(topic_queue):
-        clients = {}
-        decode = {}
+        consumers = {}
         control = {}
         control_dispose = {}
+        topic_last_offset = {}
 
         def on_next_control(obv, i):
             nonlocal control
@@ -78,58 +87,88 @@ def run_consumer(loop, source_observer, server, topics):
             print("on_next_control: {}".format(i))
 
         try:
-            while True:
-                if len(clients) == 0 or not topic_queue.empty():
-                    cmd = await topic_queue.get()
-                    if cmd[0] == 'add':
-                        print('run consumer: add')
-                        client = AIOKafkaConsumer(
-                            cmd[1],
-                            loop=loop,
-                            bootstrap_servers=server,
-                            group_id=cmd[2],
-                            auto_offset_reset='earliest')
-                        await client.start()
+            client = AIOKafkaConsumer(
+                loop=loop,
+                bootstrap_servers=server,
+                group_id=group,
+                auto_offset_reset='earliest',
+                enable_auto_commit=False
+            )
+            await client.start()
 
-                        if cmd[5] is not None:
-                            control_dispose[cmd[4]] = cmd[5].subscribe(
-                                on_next=functools.partial(on_next_control, cmd[4]),
+            while True:
+                if len(consumers) == 0 or not topic_queue.empty():
+                    cmd = await topic_queue.get()
+                    if type(cmd) is AddConsumerCmd:
+                        print('run consumer: add {}'.format(cmd.consumer.topic))
+
+                        if cmd.observer in consumers:
+                            source_observer.on_error(ValueError(
+                                "topic already subscribed for this consumer: {}".format(cmd.consumer.decode))
+                            )
+                            break
+
+                        if cmd.consumer.control is not None:
+                            control_dispose[cmd.observer] = cmd.consumer.control.subscribe(
+                                on_next=functools.partial(on_next_control, cmd.observer),
                                 on_error=source_observer.on_error,
                             )
                         print("started")
-                        if cmd[4] in clients:
-                            source_observer.on_error(ValueError(
-                                "topic already subscribed for this consumer: {}".format(cmd[3])))
-                        else:
-                            clients[cmd[4]] = client
-                            decode[cmd[4]] = cmd[3]
-                    elif cmd[0] == 'del':
-                        print('run consumer: del')
-                        dispose = control_dispose.pop(cmd[1], None)
+
+                        consumers[cmd.observer] = ConsumerProperties(
+                            observer=cmd.observer,
+                            topic=cmd.consumer.topic, decode=cmd.consumer.decode,
+                            start_from=cmd.consumer.start_from,
+                        )
+                        topic_list = set([t[1].topic for t in consumers.items()])
+                        client.subscribe(topics=topic_list)
+
+                    elif type(cmd) is DelConsumerCmd:
+                        dispose = control_dispose.pop(cmd.observer, None)
                         if dispose is not None:
                             dispose()
 
-                        client = clients.pop(cmd[1], None)
-                        if client is not None:
-                            await client.stop()
-                            cmd[1].on_completed()
+                        consumer = consumers.pop(cmd.observer, None)
+                        if consumer is not None:
+                            topic_list = set([t[1].topic for t in consumers.items()])
+                            if len(topic_list) > 0:
+                                client.subscribe(topics=topic_list)
+                            cmd.observer.on_completed()
                     else:
                         source_observer.on_error(TypeError(
                             "invalid type for queue command: {}".format(cmd)))
 
-                if len(clients) == 0:
-                    print("no client")
+                if len(consumers) == 0:
+                    print("no consumer")
                     break
 
-                for observer, client in clients.items():
-                    break  # take first entry for now
-                if observer in control:
-                    await asyncio.sleep(control[observer])
-                else:
-                    await asyncio.sleep(0)
+                regulated = False
+                for observer, consumer in consumers.items():                    
+                    if observer in control:
+                        await asyncio.sleep(control[observer])
+                        regulated = True
+                        break  # limitation only one controllable topic for now
+                
+                if regulated is False:
+                    await asyncio.sleep(0)                    
+
                 msg = await client.getone()
-                msg = decode[observer](msg.value)
-                observer.on_next(msg)
+                start_from = None
+                for observer, consumer in consumers.items():
+                    if consumer.topic == msg.topic:
+                        start_from = consumer.start_from
+                        decoded_msg = consumer.decode(msg.value)
+                        observer.on_next(decoded_msg)
+
+                tp = aiokafka.TopicPartition(msg.topic, msg.partition)
+                if start_from == 'end':                    
+                    await client.commit({tp: msg.offset + 1})
+                elif start_from == 'last':
+                    if msg.topic in topic_last_offset:
+                        await client.commit(topic_last_offset[msg.topic])
+                        topic_last_offset[msg.topic] = {tp: msg.offset + 1}
+
+            await client.stop()
 
         except asyncio.CancelledError as e:
             print("cancelled {}".format(e))
@@ -142,22 +181,20 @@ def run_consumer(loop, source_observer, server, topics):
      The kafka consumer is stated when the application subscribes to the
      create observable, and stopped on disposal
     '''
+    def on_subscribe(i, observer, scheduler):
+        print("topic subscribe: {}".format(i))
+
+        def dispose():
+            topic_queue.put_nowait(DelConsumerCmd(observer))
+
+        topic_queue.put_nowait(AddConsumerCmd(observer, i))
+        return Disposable(dispose)
+
     def on_next(i):
         print("consumer topic: {}".format(i))
-
-        def on_subscribe(decode, observer, scheduler):
-            print("topic subscribe")
-
-            def dispose():
-                topic_queue.put_nowait(('del', observer))
-
-            topic_queue.put_nowait(('add', i.topic, i.group, decode, observer, i.control))
-            return Disposable(dispose)
-
-        print(i)
         source_observer.on_next(ConsumerRecords(
             topic=i.topic,
-            records=rx.create(functools.partial(on_subscribe, i.decode))))
+            records=rx.create(functools.partial(on_subscribe, i))))
 
     task = loop.create_task(_run_consumer(topic_queue))
     topics.subscribe(
@@ -168,12 +205,13 @@ def run_consumer(loop, source_observer, server, topics):
     return task
 
 
-def run_producer(loop, source_observer, server, topics, acks, get_feedback_observer):
+def run_producer(loop, source_observer, server, topics, acks, max_request_size, get_feedback_observer):
     async def _run_producer(records):
         client = AIOKafkaProducer(
             loop=loop,
             bootstrap_servers=server,
-            acks=acks)
+            acks=acks,
+            max_request_size=max_request_size)
         pending_records = []
 
         await client.start()
@@ -237,10 +275,13 @@ def make_driver(loop=None):
             def on_next(i):
                 if type(i) is Consumer:
                     print("consumer: {}".format(i))
-                    task = run_consumer(loop, observer, i.server, i.topics)
+                    task = run_consumer(loop, observer, i.server, i.group, i.topics)
                     consumer_tasks.append(task)
                 elif type(i) is Producer:
-                    run_producer(loop, observer, i.server, i.topics, i.acks, get_feedback_observer)
+                    run_producer(
+                        loop, observer,
+                        i.server, i.topics, i.acks, i.max_request_size,
+                        get_feedback_observer)
                 else:
                     e = "Unknown item type: {}".format(i)
                     print(e)
